@@ -3,14 +3,16 @@
 # Picard, the next-generation MusicBrainz tagger
 #
 # Copyright (C) 2006-2007, 2014, 2017 Lukáš Lalinský
-# Copyright (C) 2008, 2014, 2019-2020 Philipp Wolfer
+# Copyright (C) 2008, 2014, 2019-2021 Philipp Wolfer
 # Copyright (C) 2012, 2017 Wieland Hoffmann
 # Copyright (C) 2012-2014 Michael Wiencek
-# Copyright (C) 2013-2016, 2018-2019 Laurent Monin
+# Copyright (C) 2013-2016, 2018-2021 Laurent Monin
 # Copyright (C) 2016 Suhas
 # Copyright (C) 2016-2018 Sambhav Kothari
 # Copyright (C) 2017 Sophist-UK
 # Copyright (C) 2018 Vishal Choudhary
+# Copyright (C) 2020-2021 Gabriel Ferreira
+# Copyright (C) 2021 Bob Swift
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -27,9 +29,14 @@
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
 
+from collections import defaultdict
+import inspect
 from operator import itemgetter
 import os
 import shutil
+import threading
+
+import fasteners
 
 from PyQt5 import QtCore
 
@@ -39,15 +46,21 @@ from picard import (
     PICARD_VERSION,
     log,
 )
-from picard.util import LockableObject
+from picard.profile import UserProfileGroups
 from picard.version import Version
 
 
+class Memovar:
+    def __init__(self):
+        self.dirty = True
+        self.value = None
+
+
 class ConfigUpgradeError(Exception):
     pass
 
 
-class ConfigSection(LockableObject):
+class ConfigSection(QtCore.QObject):
 
     """Configuration section."""
 
@@ -56,17 +69,11 @@ class ConfigSection(LockableObject):
         self.__qt_config = config
         self.__name = name
         self.__prefix = self.__name + '/'
-        self.__prefix_len = len(self.__prefix)
-        self._memoization = {}
+        self._memoization = defaultdict(Memovar)
 
     def key(self, name):
         return self.__prefix + name
 
-    def _subkeys(self):
-        for key in self.__qt_config.allKeys():
-            if key[:self.__prefix_len] == self.__prefix:
-                yield key[self.__prefix_len:]
-
     def __getitem__(self, name):
         opt = Option.get(self.__name, name)
         if opt is None:
@@ -74,28 +81,25 @@ class ConfigSection(LockableObject):
         return self.value(name, opt, opt.default)
 
     def __setitem__(self, name, value):
-        self.lock_for_write()
-        try:
-            key = self.key(name)
-            self.__qt_config.setValue(key, value)
-            if key in self._memoization:
-                self._memoization[key][0] = False
-        finally:
-            self.unlock()
+        key = self.key(name)
+        self.__qt_config.setValue(key, value)
+        self._memoization[key].dirty = True
 
     def __contains__(self, name):
         return self.__qt_config.contains(self.key(name))
 
+    def as_dict(self):
+        return {key: self[key] for section, key in Option.registry if section == self.__name}
+
     def remove(self, name):
-        self.lock_for_write()
+        key = self.key(name)
+        config = self.__qt_config
+        if config.contains(key):
+            config.remove(key)
         try:
-            if name in self:
-                key = self.key(name)
-                self.__qt_config.remove(key)
-                if key in self._memoization:
-                    del self._memoization[key]
-        finally:
-            self.unlock()
+            del self._memoization[key]
+        except KeyError:
+            pass
 
     def raw_value(self, name, qtype=None):
         """Return an option value without any type conversion."""
@@ -110,51 +114,160 @@ class ConfigSection(LockableObject):
         """Return an option value converted to the given Option type."""
         if name in self:
             key = self.key(name)
-            try:
-                valid, value = self._memoization[key]
-            except KeyError:
-                valid = False
+            memovar = self._memoization[key]
 
-            if not valid:
-                self.lock_for_read()
+            if memovar.dirty:
                 try:
                     value = self.raw_value(name, qtype=option_type.qtype)
                     value = option_type.convert(value)
-                    self._memoization[key] = [True, value]
+                    memovar.dirty = False
+                    memovar.value = value
                 except Exception as why:
                     log.error('Cannot read %s value: %s', self.key(name), why, exc_info=True)
                     value = default
-                self.unlock()
-            return value
+                return value
+            else:
+                return memovar.value
         return default
 
 
+class SettingConfigSection(ConfigSection):
+    """Custom subclass to automatically accommodate saving and retrieving values based on user profile settings.
+    """
+    PROFILES_KEY = 'user_profiles'
+    SETTINGS_KEY = 'user_profile_settings'
+
+    @classmethod
+    def init_profile_options(cls):
+        ListOption.add_if_missing("profiles", cls.PROFILES_KEY, [])
+        Option.add_if_missing("profiles", cls.SETTINGS_KEY, {})
+
+    def __init__(self, config, name):
+        super().__init__(config, name)
+        self.__qt_config = config
+        self.__name = name
+        self.__prefix = self.__name + '/'
+        self._memoization = defaultdict(Memovar)
+        self.init_profile_options()
+        self.profiles_override = None
+        self.settings_override = None
+
+    def _get_active_profile_ids(self):
+        if self.profiles_override is None:
+            profiles = self.__qt_config.profiles[self.PROFILES_KEY]
+        else:
+            profiles = self.profiles_override
+        if profiles is None:
+            return
+        for profile in profiles:
+            if profile['enabled']:
+                yield profile["id"]
+
+    def _get_active_profile_settings(self):
+        for id in self._get_active_profile_ids():
+            yield id, self._get_profile_settings(id)
+
+    def _get_profile_settings(self, id):
+        if self.settings_override is None:
+            profile_settings = self.__qt_config.profiles[self.SETTINGS_KEY][id]
+        else:
+            profile_settings = self.settings_override[id]
+        if profile_settings is None:
+            log.error("Unable to find settings for user profile '%s'", id)
+            return {}
+        return profile_settings
+
+    def __getitem__(self, name):
+        # Don't process settings that are not profile-specific
+        if name in UserProfileGroups.get_all_settings_list():
+            for id, settings in self._get_active_profile_settings():
+                if name in settings and settings[name] is not None:
+                    return settings[name]
+        opt = Option.get(self.__name, name)
+        if opt is None:
+            return None
+        return self.value(name, opt, opt.default)
+
+    def __setitem__(self, name, value):
+        # Don't process settings that are not profile-specific
+        if name in UserProfileGroups.get_all_settings_list():
+            for id, settings in self._get_active_profile_settings():
+                if name in settings:
+                    self._save_profile_setting(id, name, value)
+                    return
+        key = self.key(name)
+        self.__qt_config.setValue(key, value)
+        self._memoization[key].dirty = True
+
+    def _save_profile_setting(self, profile_id, name, value):
+        profile_settings = self.__qt_config.profiles[self.SETTINGS_KEY]
+        profile_settings[profile_id][name] = value
+        key = self.__qt_config.profiles.key(self.SETTINGS_KEY)
+        self.__qt_config.setValue(key, profile_settings)
+        self._memoization[key].dirty = True
+
+    def set_profiles_override(self, new_profiles=None):
+        self.profiles_override = new_profiles
+
+    def set_settings_override(self, new_settings=None):
+        self.settings_override = new_settings
+
+
 class Config(QtCore.QSettings):
 
-    """Configuration."""
+    """Configuration.
+    QSettings is not thread safe, each thread must use its own instance of this class.
+    Use `get_config()` to obtain a Config instance for the current thread.
+    Changes to one Config instances are automatically available to all other instances.
+
+    Use `Config.from_app` or `Config.from_file` to obtain a new `Config` instance.
+
+    See: https://doc.qt.io/qt-5/qsettings.html#accessing-settings-from-multiple-threads-or-processes-simultaneously
+    """
 
     def __init__(self):
-        self.__known_keys = set()
+        # Do not call `QSettings.__init__` here. The proper overloaded `QSettings.__init__`
+        # gets called in `from_app` or `from_config`. Only those class methods must be used
+        # to create a new instance of `Config`.
+        pass
 
     def __initialize(self):
         """Common initializer method for :meth:`from_app` and
         :meth:`from_file`."""
 
+        self.setAtomicSyncRequired(False)  # See comment in event()
         self.application = ConfigSection(self, "application")
-        self.setting = ConfigSection(self, "setting")
+        self.profiles = ConfigSection(self, "profiles")
+        self.setting = SettingConfigSection(self, "setting")
         self.persist = ConfigSection(self, "persist")
-        self.profile = ConfigSection(self, "profile/default")
-        self.current_preset = "default"
 
         TextOption("application", "version", '0.0.0dev0')
         self._version = Version.from_string(self.application["version"])
         self._upgrade_hooks = dict()
-        # Save known config names for faster access and to prevent
-        # strange cases of Qt locking up when accessing QSettings.allKeys()
-        # or QSettings.contains() shortly after having written settings
-        # inside threads.
-        # See https://tickets.metabrainz.org/browse/PICARD-1590
-        self.__known_keys = set(self.allKeys())
+
+    def event(self, event):
+        if event.type() == QtCore.QEvent.UpdateRequest:
+            # Syncing the config file can trigger a deadlock between QSettings internal mutex and
+            # the Python GIL in PyQt up to 5.15.2. Workaround this by handling this ourselves
+            # with custom file locking.
+            # See also https: // tickets.metabrainz.org/browse/PICARD-2088
+            log.debug('Config file update requested on thread %r', threading.get_ident())
+            self.sync()
+            return True
+        else:
+            return super().event(event)
+
+    def sync(self):
+        # Custom file locking for save multi process syncing of the config file. This is needed
+        # as we have atomicSyncRequired disabled.
+        with fasteners.InterProcessLock(self.get_lockfile_name()):
+            super().sync()
+
+    def get_lockfile_name(self):
+        filename = self.fileName()
+        directory = os.path.dirname(filename)
+        filename = '.' + os.path.basename(filename) + '.synclock'
+        return os.path.join(directory, filename)
 
     @classmethod
     def from_app(cls, parent):
@@ -192,27 +305,6 @@ class Config(QtCore.QSettings):
         this.__initialize()
         return this
 
-    def setValue(self, key, value):
-        super().setValue(key, value)
-        self.__known_keys.add(key)
-
-    def remove(self, key):
-        super().remove(key)
-        self.__known_keys.discard(key)
-
-    def contains(self, key):
-        # Overwritten due to https://tickets.metabrainz.org/browse/PICARD-1590
-        # See comment above in __initialize
-        return key in self.__known_keys or super().contains(key)
-
-    def switchProfile(self, profilename):
-        """Sets the current profile."""
-        key = "profile/%s" % (profilename,)
-        if self.contains(key):
-            self.profile.name = key
-        else:
-            raise KeyError("Unknown profile '%s'" % (profilename,))
-
     def register_upgrade_hook(self, func, *args):
         """Register a function to upgrade from one config version to another"""
         to_version = Version.from_string(func.__name__)
@@ -225,6 +317,11 @@ class Config(QtCore.QSettings):
 
     def run_upgrade_hooks(self, outputfunc=None):
         """Executes registered functions to upgrade config version to the latest"""
+        if self._version == Version(0, 0, 0, 'dev', 0):
+            # This is a freshly created config
+            self._version = PICARD_VERSION
+            self._write_version()
+            return
         if not self._upgrade_hooks:
             return
         if self._version >= PICARD_VERSION:
@@ -296,18 +393,40 @@ class Option(QtCore.QObject):
     qtype = None
 
     def __init__(self, section, name, default):
+        key = (section, name)
+        if key in self.registry:
+            stack = inspect.stack()
+            fmt = "Option %s/%s already declared"
+            args = [section, name]
+            if len(stack) > 1:
+                f = stack[1]
+                fmt += "\nat %s:%d: in %s"
+                args.extend((f.filename, f.lineno, f.function))
+                if f.code_context:
+                    fmt += "\n%s"
+                    args.append("\n".join(f.code_context).rstrip())
+            log.error(fmt, *args)
         super().__init__()
         self.section = section
         self.name = name
         self.default = default
         if not hasattr(self, "convert"):
             self.convert = type(default)
-        self.registry[(self.section, self.name)] = self
+        self.registry[key] = self
 
     @classmethod
     def get(cls, section, name):
         return cls.registry.get((section, name))
 
+    @classmethod
+    def add_if_missing(cls, section, name, default):
+        if not cls.exists(section, name):
+            cls(section, name, default)
+
+    @classmethod
+    def exists(cls, section, name):
+        return (section, name) in cls.registry
+
 
 class TextOption(Option):
 
@@ -340,13 +459,23 @@ class ListOption(Option):
 config = None
 setting = None
 persist = None
+profiles = None
 
 
-def _setup(app, filename=None):
-    global config, setting, persist
+def setup_config(app, filename=None):
+    global config, setting, persist, profiles
     if filename is None:
         config = Config.from_app(app)
     else:
         config = Config.from_file(app, filename)
     setting = config.setting
     persist = config.persist
+    profiles = config.profiles
+
+
+def get_config():
+    """Returns a config object for the current thread.
+
+    Config objects for threads are created on demand and cached for later use.
+    """
+    return config
